# Generated by Django 4.2.16 on 2024-10-30 23:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContractExclusionSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(blank=True, null=True)),
                ('is_active', models.BooleanField(default=True)),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.AddField(
            model_name='contract',
            name='contract_exclusion_set_id',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name='ContractExclusionSetExcludedWorkType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contract_exclusion_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.contractexclusionset')),
                ('excluded_work_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.billingcode')),
            ],
        ),
        migrations.CreateModel(
            name='ContractExclusionSetExcludedRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contract_exclusion_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.contractexclusionset')),
                ('excluded_role', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.role')),
            ],
        ),
        migrations.AddField(
            model_name='contractexclusionset',
            name='excluded_roles',
            field=models.ManyToManyField(related_name='excluded_role_sets', through='djautotask.ContractExclusionSetExcludedRole', to='djautotask.role'),
        ),
        migrations.AddField(
            model_name='contractexclusionset',
            name='excluded_work_types',
            field=models.ManyToManyField(related_name='excluded_work_type_sets', through='djautotask.ContractExclusionSetExcludedWorkType', to='djautotask.billingcode'),
        ),
        migrations.CreateModel(
            name='ContractExcludedRoleTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionsetrole',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionsetexcludedrole',),
        ),
        migrations.CreateModel(
            name='ContractExcludedWorkTypeTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionsetworktype',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionsetexcludedworktype',),
        ),
        migrations.CreateModel(
            name='ContractExclusionSetTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionset',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionset',),
        ),
    ]
