# Generated by Django 4.2.16 on 2024-11-01 13:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0118_contractexclusionset_and_more'),
        ('djautotask', '0119_merge_20241024_1521'),
    ]

    operations = [
    ]
