# Generated by Django 4.2.16 on 2024-10-08 02:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='tasktypetracker',
            table=None,
        ),
        migrations.CreateModel(
            name='CompanyAlert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('alert_text', models.TextField(blank=True, max_length=8000, null=True)),
                ('alert_type', models.IntegerField(blank=True, null=True)),
                ('account', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='alerts', to='djautotask.account')),
            ],
        ),
        migrations.CreateModel(
            name='CompanyAlertTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_companyalert',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.companyalert',),
        ),
    ]
