# Generated by Django 2.1.14 on 2020-01-31 16:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0039_merge_20200131_1022'),
        ('djautotask', '0039_auto_20200131_1134'),
    ]

    operations = [
    ]
