# Generated by Django 4.2.16 on 2024-10-24 15:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0118_alter_tasktypetracker_table_companyalert_and_more'),
        ('djautotask', '0118_alter_ticket_due_date_time'),
    ]

    operations = [
    ]
