# Generated by Django 4.2.16 on 2024-10-23 14:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='due_date_time',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]
