# Generated by Django 4.0.7 on 2023-03-01 13:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0111_alter_phase_estimated_hours'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='owner_resource',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.resource'),
        ),
    ]
