# Generated by Django 2.1.14 on 2019-11-29 15:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0031_auto_20191129_1454'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='displaycolor',
            options={'ordering': ('label',), 'verbose_name_plural': 'Display colors'},
        ),
        migrations.AlterModelOptions(
            name='projectstatus',
            options={'ordering': ('label',), 'verbose_name_plural': 'Project statuses'},
        ),
        migrations.AlterModelOptions(
            name='queue',
            options={'ordering': ('label',), 'verbose_name_plural': 'Queues'},
        ),
        migrations.AlterModelOptions(
            name='status',
            options={'ordering': ('label',), 'verbose_name_plural': 'Statuses'},
        ),
    ]
