# Generated by Django 2.1.11 on 2020-04-16 12:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0060_auto_20200414_1522'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='servicecallstatus',
            options={'ordering': ('label',), 'verbose_name_plural': 'Service call statuses'},
        ),
    ]
