# Generated by Django 4.0 on 2022-02-17 15:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0103_auto_20220106_1622'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AllocationCodeTracker',
        ),
        migrations.RenameModel(
            old_name='AllocationCode',
            new_name='BillingCode',
        ),
        migrations.CreateModel(
            name='BillingCodeTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_billingcode',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.billingcode',),
        ),
        migrations.RenameField(
            model_name='task',
            old_name='allocation_code',
            new_name='billing_code',
        ),
        migrations.RenameField(
            model_name='ticket',
            old_name='allocation_code',
            new_name='billing_code',
        ),
        migrations.RenameField(
            model_name='timeentry',
            old_name='allocation_code',
            new_name='billing_code',
        ),
    ]
