# Generated by Django 2.1.14 on 2020-07-14 08:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0069_auto_20200602_1419'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queue',
            name='label',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]
