# Generated by Django 4.2.11 on 2024-06-17 10:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0115_task_task_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicecall',
            name='tasks',
            field=models.ManyToManyField(through='djautotask.ServiceCallTask', to='djautotask.task'),
        ),
        migrations.AlterField(
            model_name='servicecall',
            name='tickets',
            field=models.ManyToManyField(through='djautotask.ServiceCallTicket', to='djautotask.ticket'),
        ),
    ]
